<?php

namespace InSegment\ApiCore\Services;

use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Schema;

use InSegment\ApiCore\Facades\DTODefs;
use InSegment\ApiCore\Exceptions\ApiTransactorException;
use InSegment\ApiCore\Scopes\TransactionSearchScope;
use InSegment\ApiCore\Models\DeletedTransactionRecord;
use InSegment\ApiCore\Models\LogReceivedRecord;
use InSegment\ApiCore\Models\UUIDGeneration;
use InSegment\ApiCore\Models\Transaction;
use InSegment\ApiCore\Services\ExceptionDebug;
use InSegment\ApiCore\Api\DataTransferObject;

use Illuminate\Database\Schema\Blueprint;
use Illuminate\Database\Eloquent\Model;

/**
 * Class Transactor
 *
 * @package App
 */
final class Transactor
{
    /**
     * An instance
     * 
     * @var \InSegment\ApiCore\Services\Transactor 
     */
    private static $instance = null;
    
    /**
     * An array holding state of Transaction
     * 
     * @var array 
     */
    private $transactionData = [];
    
    /**
     * An array holding all created temporary tables
     * 
     * @var array 
     */
    private $temporaryTables = [];
    
    /**
     * Show create parser
     * 
     * @var \InSegment\ApiCore\Services\ParseCreateTable 
     */
    private $showCreateParser;
    
    /**
     * An array of ids for existence in update table for each table for table type
     * 
     * @var array 
     */
    private $existence = ['insert' => [], 'update' => []];
    
    /**
     * Session identifier
     * 
     * @var string
     */
    private $sessionId;
    
    /**
     * Uuidables
     * 
     * @var array 
     */
    private $uuidables;
    
    /**
     * Reserved UUIDs by table
     * 
     * @var array 
     */
    private $uuidReserve = [];
    
    /**
     * Buffer table name
     * 
     * @var array 
     */
    private $bufferTable = [];
    
    /**
     * BufferedWriter instance
     * 
     * @var \InSegment\ApiCore\Services\BufferedWriter
     */
    private $bufferedWriter;
    
    /**
     * Results of static inspection of rules
     * 
     * @var array 
     */
    private $inspectionResults;
    
    /**
     * Allowed transaction tables
     * 
     * @var array
     */
    private static $allowedTransactionTables;
    
    /**
     * Allowed transaction tables key names
     * 
     * @var array
     */
    private static $allowedTransactionTablesKeys;
    
    /**
     * Tables which enabled deletion from them
     * 
     * @var array 
     */
    private static $deletionEnabledTables;
    
    /**
     * If Transactor is currently active
     * 
     * @var bool
     */
    private static $isWatching = false;
    
    /**
     * Only those tables are available to open transactions for
     * 
     * @param bool $readOnly
     * @param string $rootClass
     * @return array
     */
    public static function getAllowedTransactionTables(bool $readOnly = false, string $rootClass = null) : array
    {
        $workLoad = ['ew-with-read-only' => $readOnly];

        if (!isset($rootClass)) {
            if (!isset(self::$allowedTransactionTables)) {
                self::getInstance()->runInspections($workLoad);
            }
            
            return self::$allowedTransactionTables;
        }
        
        return DTODefs::instance()->performInspection($rootClass, 'enabled-write', $workLoad)['enabled'];
    }
    
    /**
     * Only those tables keys are available to open transactions for
     * 
     * @return array
     */
    public static function getAllowedTransactionTablesKeys() : array
    {
        if (!isset(self::$allowedTransactionTablesKeys)) {
            $workLoad = ['ew-with-read-only' => false];
            self::getInstance()->runInspections($workLoad);
        }
        
        return self::$allowedTransactionTablesKeys;
    }
    
    /**
     * Only those tables which enabled deletion from them during transact
     * 
     * @param bool $readOnly
     * @param string $rootClass
     * @return array
     */
    public static function getDeletionEnabledTables(bool $readOnly = false, string $rootClass = null) : array
    {
        $workLoad = ['ew-with-read-only' => $readOnly];

        if (!isset($rootClass)) {
            if (isset(self::$deletionEnabledTables)) {
                return self::$deletionEnabledTables;
            }

            self::getInstance()->runInspections($workLoad);
            return self::$deletionEnabledTables;
        }
        
        return DTODefs::instance()->performInspection($rootClass, 'enabled-write', $workLoad)['deletable'];
    }
    
    /**
     * Get transaction UID
     * 
     * @return type
     */
    public static function getTransctionUID()
    {
        $instance = self::getInstance();
        if (!isset($instance->transactionData['UID'])) {
            self::requireTransactionState();
            throw (new ApiTransactorException(ApiTransactorException::CODE_UNABLE_TO_GENERATE_UID))->compile();
        }
        
        return $instance->transactionData['UID'];
    }
    
    /**
     * Lists column unique data
     * 
     * @param string $rootClass
     * @return array
     */
    public static function getColumnUniqueData(string $rootClass): array
    {
        $instance = self::getInstance();
        
        foreach (self::getAllowedTransactionTables(true, $rootClass) as $modelClass => $table) {
            $table = (new $modelClass)->getTable();
            
            $external = DTODefs::getExternalKeyName($modelClass, true);
            
            $instance->showCreateParser->loadColumnData($table, $external);
        }
        
        return $instance->showCreateParser->getColumnUniqueData();
    }
    
    /**
     * Get unique keys which are using specified column(s) of the specified model
     * 
     * @param string $modelClass
     * @param array $columns
     * @return array
     */
    public static function getUniquesForColumns(string $modelClass, array $columns): array
    {
        $instance = self::getInstance();
        $table = (new $modelClass)->getTable();
        $instance->showCreateParser->loadColumnData($table);
        $uniqueData = &$instance->showCreateParser->getColumnUniqueData();
        
        if (!isset($uniqueData[$table])) {
            return []; // data for this model's uniques is not found
        }
        
        $results = [];
        
        foreach ($uniqueData[$table] as $unique) {
            if (count(array_intersect($columns, $unique))) {
                $results[] = $unique;
            }
        }
        
        return $results;
    }
    
    /**
     * Get enabled eager loads list
     * 
     * @param string $rootClass
     * @param array $categories
     * @return array
     */
    public static function getEnabledEagers(string $rootClass, array $categories = ['all', 'export']): array
    {
        $instance = self::getInstance();
        !isset($instance->inspectionResults['eagers']) && $instance->runInspections([], $rootClass);
        $inspection = $instance->inspectionResults['eagers'];
        
        if (self::isWatching()) {
            // nullify by reference logic
            foreach ($inspection['tables'] as $table => &$relationPathes) {
                if (!isset($instance->temporaryTables[$table])) {
                    foreach ($relationPathes as &$path) {
                        $path = null;
                    }
                }
            }
        }
        
        $eagers = array_values(array_unique(array_filter(
            (in_array('all', $categories) ? array_get($inspection, 'relations.all', []) : []) +
            (in_array('export', $categories) ? array_get($inspection, 'relations.export', []) : [])
        )));
        
        return $eagers;
    }
    
    /**
     * Get preload Closure for $class, if any
     * 
     * @param string $classInQuestion
     * @param string $extKeysImplode
     * @return \Closure|null
     */
    public static function getPreloads(string $classInQuestion, string $extKeysImplode)
    {
        $instance = self::getInstance();
        !isset($instance->inspectionResults['preload']) && $instance->runInspections();
        return $instance->inspectionResults['preload']['suppliers'][$classInQuestion][$extKeysImplode] ?? null;
    }
    
    /**
     * Use input data to fulfill wildcards for preload
     * 
     * @param array $data
     * @return null
     */
    public static function fulfilReferenceWildcards(&$data)
    {
        $instance = self::getInstance();
        !isset($instance->inspectionResults['preload']) && $instance->runInspections();
        
        foreach (array_get($instance->inspectionResults, 'preload.referenceWildcards', []) as $class => $wildcards) {
            foreach ($wildcards as $externalKey => $wildcard) {
                $array = is_array($wildcard);
                $count = $array ? count($wildcard) : 1;
                
                if ($count > 1) {
                    $segments = [];
                    foreach ($wildcard as $k => $wildCardPart) {
                        $segments[] = data_get($data, $wildCardPart, []);
                    }
                    
                    $out = [];
                    foreach ($segments as &$segment) {
                        foreach ($segment as $key => &$value) {
                            $out[$key][] = &$value;
                        }
                    }
                    
                    $instance->inspectionResults['preload']['reference'][$class][$externalKey] = $out;
                } else {
                    if ($array) {
                        $wildcard = $wildcard[0];
                    }
                    
                    $instance->inspectionResults['preload']['reference'][$class][$externalKey] = data_get($data, $wildcard, []);
                }
            }
        }
    }
    
    /**
     * Method to be called by global exception handler to transform any exceptions happened during active transaction
     * 
     * @param \Throwable $exception
     * @return \Throwable
     */
    public static function affectOnYourWatch(\Throwable $exception): \Throwable
    {
        if (!self::$isWatching || self::isNotPrepared() || $exception instanceof ApiTransactorException) {
            return $exception;
        }
        
        if (!env('API_TRANSACTION_DEBUG')) {
            $instance = self::getInstance();
            $instance->transactionData['isPrepared'] = 'abnormal';
            app('session')->put(['transaction' => $instance->transactionData]);
        }
        
        return new ApiTransactorException(ApiTransactorException::CODE_SERVER_FAULT, $exception);
    }
    
    /**
     * Checks if the Transactor instance is prepared correctly
     * 
     * @return bool
     */
    public static function isPrepared()
    {
        return array_get(self::getInstance()->transactionData, 'isPrepared', 0) == 1;
    }
    
    /**
     * Checks if the Transactor instance is not prepared at all
     * 
     * @return bool
     */
    public static function isNotPrepared()
    {
        return array_get(self::getInstance()->transactionData, 'isPrepared', 0) == false;
    }

    /**
     * Get status of Transactor
     * 
     * @param array|null $columns
     * @return array
     */
    public static function status(array $columns = null)
    {
        try {
            $errMode = ['errmode' => DB::connection()->getPdo()->getAttribute(\PDO::ATTR_ERRMODE)];
        } catch (\Throwable $exception) {
            $errMode = ['errmode' => 'unknown'];
            ExceptionDebug::makeDebug($errMode, $exception, 'errmode_exception');
        }
        
        $source = self::getInstance()->transactionData + $errMode;
        
        return isset($columns) ? array_intersect_key($source, array_flip($columns)) : $source;
    }
    
    /**
     * If the Transactor transacting $table and temporary ones exist, returns true
     * 
     * @param string $modelClass A string representing Model class
     * @return boolean
     */
    public static function doesHaveTables(string $modelClass)
    {
        $instance = self::getInstance();
        $allowedTransactionTables = self::getAllowedTransactionTables();
        
        if (isset($allowedTransactionTables[$modelClass])) {
            $modelTable = $allowedTransactionTables[$modelClass];
            return isset($instance->temporaryTables[$modelTable]);
        }
        
        return false;
    }
    
    /**
     * If the Transactor transacting $table, returns temporary insert table
     * Otherwise returns null
     * 
     * @param string $modelClass A string representing Model class
     * @param bool $strict
     * @return string|null
     */
    public static function getInsertTable(string $modelClass, bool $strict = false)
    {
        $instance = self::getInstance();
        $allowedTransactionTables = self::getAllowedTransactionTables();
        
        if (isset($allowedTransactionTables[$modelClass])) {
            $modelTable = $allowedTransactionTables[$modelClass];
            if (isset($instance->temporaryTables[$modelTable])) {
                return $instance->temporaryTables[$modelTable]['insert'];
            }
        }
        
        if ($strict) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_TABLE_NOT_INCLUDED))->compile((new $modelClass)->getTable());
        }
        
        return null;
    }
    
    /**
     * Execute actions on an instance of Transactor within the state special handling for Eloquent operations on Models
     * if the transaction is prepared, or just in plain state otherwise
     * 
     * @param callable $transactionAction
     * @return mixed
     */
    public static function watchingTables(callable $transactionAction = null)
    {
        if (self::$isWatching) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_RECURSIVE_TRANSACTION))->compile();
        }
        
        // do not check foreign keys while importing
        DB::statement("SET FOREIGN_KEY_CHECKS=0");

        self::startWatchingTables();

        try {
            $result = $transactionAction(self::getInstance());
        } catch (\Throwable $e) {
            throw $e;
        } finally {
            self::stopWatchingTables();

            // re-enable foreign key checks
            DB::statement("SET FOREIGN_KEY_CHECKS=1");
            self::$isWatching = false;
        }

        return $result;
    }
    
    /**
     * Whether the Transactor is in the mode of special handling for Eloquent operations on Models
     * 
     * @return bool
     */
    public static function isWatching()
    {
        return self::$isWatching;
    }
    
    /**
     * Enable special handling for Eloquent operations on Models
     * 
     * @return null
     */
    private static function startWatchingTables()
    {
        App::make(JoinResolver::class)->enabled = true;
        
        if (self::isPrepared()) {
            self::$isWatching = true;
            
            foreach (array_keys(self::getAllowedTransactionTables()) as $class) {
                TransactionSearchScope::applyTo($class);
            }

            BufferedBuilder::$buffer = self::getInstance();
        }
    }
    
    /**
     * Disable special handling for Eloquent operations on Models
     * 
     * @return null
     */
    private static function stopWatchingTables()
    {
        if (self::isPrepared()) {
            BufferedBuilder::$buffer = null;
            TransactionSearchScope::disable();
            self::$isWatching = false;
        }
        
        App::make(JoinResolver::class)->enabled = false;
    }
    
    /**
     * Require to be prepared and not abnormal
     */
    public static function requireTransactionState()
    {
        if (self::isNotPrepared()) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_SEQUENCE))->compile('not prepared');
        } else if (!self::isPrepared()) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_FAIL_NOT_UNDONE))->compile();
        }
    }
    
    /**
     * Get the (cached) key existence of the Model instance with specified id in $type table for Model of $modelClass
     * 
     * @param string $table
     * @param string $modelClass
     * @param mixed $keyValue
     * @param string $keyName
     * @return bool
     */
    public static function getExistence(string $table, string $modelClass, $keyValue, string $keyName = 'id')
    {
        $instance = self::getInstance();
        $allowedTransactionTables = self::getAllowedTransactionTables();
        
        if (isset($allowedTransactionTables[$modelClass])) {
            $modelTable = $allowedTransactionTables[$modelClass];
            
            if (isset($instance->temporaryTables[$modelTable])) {
                if (!DTODefs::isUUIDable($modelTable) && (new $modelClass)->_incrementing) {
                    // separates already insered entries from those to update. Is a sum of old AUTO_INCREMENT and new from insert table
                    $updateSeparator = $instance->transactionData['oldAI'][$modelTable] + $instance->transactionData['new'][$modelTable];

                    if ($instance->temporaryTables[$modelTable][$type = 'insert'] == $table) {
                        if ($keyValue < $updateSeparator) {
                            throw (new ApiTransactorException(ApiTransactorException::CODE_IMPOSSIBLE_INSERT))
                                ->compile($modelClass, $keyValue, $updateSeparator);
                        }
                    } else if ($instance->temporaryTables[$modelTable][$type = 'update'] == $table) {
                        if ($keyValue >= $updateSeparator) {
                            throw (new ApiTransactorException(ApiTransactorException::CODE_IMPOSSIBLE_UPDATE))
                                ->compile($modelClass, $keyValue, $updateSeparator);
                        }
                    } else {
                        throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_TABLE_FOR_SAVE))->compile($table);
                    }
                    
                    if (!isset($instance->existence[$type][$modelTable])) {
                        self::addToExistence($instance, $type, $modelTable, $keyName);
                    }

                    return array_key_exists($keyValue, $instance->existence[$type][$modelTable]);
                } else {
                    foreach (['insert', 'update'] as $type) {
                        if (!isset($instance->existence[$type][$modelTable])) {
                            self::addToExistence($instance, $type, $modelTable, $keyName);
                        }
                        
                        if (array_key_exists($keyValue, $instance->existence[$type][$modelTable])) {
                            return true;
                        }
                    }
                }

            }
        }
        
        return false;
    }
    
    /**
     * Get the list of IDs of the objects, that were changed during the current transaction.
     * 
     * @param string $modelClass
     * @param string $keyName
     * @return array|false
     */
    public static function getTransactedChangesIds(string $modelClass, string $keyName = 'id')
    {
        $instance = self::getInstance();
        if (!($modelTable = $instance->getModelTable($modelClass))) {
            return false;
        }
        
        $result = [];

        foreach (['insert', 'update'] as $type) {
            if (!isset($instance->existence[$type][$modelTable])) {
                self::addToExistence($instance, $type, $modelTable, $keyName);
            }

            $result = array_merge($result, array_keys($instance->existence[$type][$modelTable]));
        }

        return array_unique($result);
    }
    
    /**
     * Scope of the all transacted models (including those not changed directly)
     * 
     * @param \Illuminate\Database\Eloquent\Builder $query
     * @param string $modelClass
     * @param string $keyName
     * @param bool $strict
     * @return \Illuminate\Database\Eloquent\Builder|null
     */
    public static function receivedModelsScope($query, string $keyName = null, bool $strict = false)
    {
        $model = $query->getModel();
        if (!isset($keyName)) {
            $keyName = $model->getKeyName();
        }
        
        if (($receivedIdsQuery = self::getInstance()->getReceivedIdsQuery(get_class($model), $keyName))) {
            return $query
                ->whereRaw("`{$keyName}` IN ({$receivedIdsQuery->toSql()})")
                ->mergeBindings($receivedIdsQuery->toBase());
        } else if ($strict) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_TABLE_NOT_INCLUDED))->compile($model->getTable());
        } else {
            return null;
        }
    }

    /**
     * Get or create Transaction
     * 
     * @param mixed $data Counts.
     * @return void
     */
    public static function acquire(&$data)
    {
        $rootClass = data_get($data, 'root', '');
        $rootDTOs = DTODefs::getGlobalRootDTOs();

        if (!$rootClass || !in_array($rootClass, $rootDTOs)) {
            $rootClass = $rootDTOs[0];
        }

        self::prepare($rootClass, data_get($data, 'count', []));
    }

    /**
     * Set or update Object
     * 
     * @param string $type
     * @param bool $update
     * @param mixed $data
     * @return mixed
     */
    public static function import(&$type, &$data, &$update)
    {
        $result = self::watchingTables(function (Transactor $transactor) use ($type, $data, $update)
            {
                $ret = (new DataTransferObject($type, $data, str_singular($type)))->processIncoming($update);
                $transactor->getBufferedWriter()->flushAll();
                return $ret;
            });
            
        return $result;
    }

    
    /**
     * Undo Transaction
     * 
     */
    public static function undoTransaction()
    {
        self::watchingTables(function (Transactor $transactor)
        {
            return $transactor->undo();
        });
    }
    
    /**
     * Close Transaction
     * 
     * @param bool $clear
     * @return mixed
     */
    public static function close(&$clear)
    {
        $result = self::watchingTables(function (Transactor $transactor) use ($clear) {
            $mergeOutput = DTODefs::onTransactionMerge($transactor);
            
            return $transactor->commit(function (Transactor $transactor) use ($mergeOutput) {
                return DTODefs::postProcessTransaction($transactor, $mergeOutput);
            }, $clear);
        });
        
        return $result;
    }

    /**
     * Get query for IDs of Models received in transaction (including not changed ones)
     * 
     * @param string $modelClass
     * @param string $keyName
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public function getReceivedIdsQuery(string $modelClass, string $keyName = null)
    {
        $instance = self::getInstance();
        $modelExample = (new $modelClass);
        $modelTable = $modelExample->getTable();
        if (!isset($keyName)) {
            $keyName = $modelExample->getKeyName();
        }
        
        if (!isset($instance->temporaryTables[$modelTable])) {
            return false;
        }
        
        return LogReceivedRecord
            ::select(['key'])
            ->where('table', '=', $modelTable)
            ->where('key_name', '=', $keyName);
    }
    
    /**
     * Get query for IDs of Models received in transaction (including not changed ones)
     * 
     * @param string $modelClass
     * @param string $keyName
     * @return \Illuminate\Database\Eloquent\Builder
     */
    public function getKeysForDeleteQuery(string $modelClass, string $keyName = null)
    {
        $instance = self::getInstance();
        $modelExample = (new $modelClass);
        $modelTable = $modelExample->getTable();
        if (!isset($keyName)) {
            $keyName = $modelExample->getKeyName();
        }
        
        if (!isset($instance->temporaryTables[$modelTable])) {
            return false;
        }
        
        return DeletedTransactionRecord
            ::select(['key'])
            ->where('table', '=', $modelTable)
            ->where('key_name', '=', $keyName);
    }
    
    /**
     * Get BufferedWriter instance
     * 
     * @return \InSegment\ApiCore\Services\BufferedWriter
     */
    public function getBufferedWriter()
    {
        return $this->bufferedWriter;
    }
    
    /**
     * Get table for model class
     * 
     * @param string $modelClass
     * @return string|boolean
     */
    protected function getModelTable(string $modelClass)
    {
        $allowedTransactionTables = self::getAllowedTransactionTables();

        if (!isset($allowedTransactionTables[$modelClass])) {
            return false;
        }

        $modelTable = $allowedTransactionTables[$modelClass];
        
        if (!isset($this->temporaryTables[$modelTable])) {
            return false;
        }

        return $modelTable;
    }

    /**
     * Populate Transactor->existence property with a new value.
     * 
     * @param \InSegment\ApiCore\Services\Transactor $instance
     * @param string $type
     * @param string $modelTable
     * @param string $keyName
     */
    protected static function addToExistence(Transactor $instance, string $type, string $modelTable, string $keyName)
    {
        $instance->existence[$type][$modelTable] = array_flip(
            DB::table($instance->temporaryTables[$modelTable][$type])
                ->pluck($keyName)
                ->toArray()
        );
    }
    
    /**
     * If the Transactor transacting $table, returns temporary update table
     * Otherwise returns null
     * 
     * @param string $modelClass A string representing Model class
     * @param bool $strict
     * @return string|null
     */
    public static function getUpdateTable(string $modelClass, bool $strict = false)
    {
        $instance = self::getInstance();
        $allowedTransactionTables = self::getAllowedTransactionTables();
        
        if (isset($allowedTransactionTables[$modelClass])) {
            $modelTable = $allowedTransactionTables[$modelClass];
            if (isset($instance->temporaryTables[$modelTable])) {
                return $instance->temporaryTables[$modelTable]['update'];
            }
        }
        
        if ($strict) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_TABLE_NOT_INCLUDED))->compile((new $modelClass)->getTable());
        }
        
        return null;
    }
    
    /**
     * Get session object of the specified class
     * 
     * @param string $class
     * @return mixed
     */
    public static function getSessionObject(string $class) {
        $instance = self::getInstance();
        return array_get($instance->transactionData['objects'], $class);
    }
    
    /**
     * Get session object of the specified class
     * 
     * @param string $class
     * @param mixed $object
     * @return null
     */
    public static function setSessionObject(string $class, $object = null) {
        $instance = self::getInstance();
        if ($object == null) {
            unset($instance->transactionData['objects'][$class]);
        } else {
            $instance->transactionData['objects'][$class] = $object;
        }
        app('session')->put(['transaction' => $instance->transactionData]);
    }
    
    /**
     * Attempt to commit the transaction effects into the actual db
     * 
     * @param \Closure|null $postProcessing
     * @param bool $clear
     * @return array result of commit (empty array on success)
     * @throws \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    public function commit(\Closure $postProcessing = null, bool $clear = true)
    {
        if (!self::isPrepared()) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_SEQUENCE))->compile('not prepared');
        }
        
        /**
         * Check that all-right, and we received what we wanted in sufficient amount
         */
        $checks = [
            'a_i' => array_keys($this->transactionData['oldAI']),
            'reserved' => array_keys($this->transactionData['reserved'] ?? [])
        ];
        
        foreach ($checks as $check => $tables) {
            foreach ($tables as $table) {
                if ($check == 'a_i') {
                    $oldAI = $this->transactionData['oldAI'][$table];
                    $newAI = $this->transactionData['newAI'][$table];
                    $received = $this->transactionData['received'][$table];
                    $val1 = $oldAI + $received;
                    $val2 = $newAI;
                } else if ($check == 'reserved') {
                    $val1 = $this->transactionData['received'][$table];
                    $val2 = $this->transactionData['reserved'][$table];
                }

                if ($val1 < $val2) {
                    if (empty($this->transactionData['dynUW'][$table])) {
                        throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_WEIGHT))
                            ->compile('Transaction is underweight')->setData(compact('table'));
                    }
                } else if ($val1 > $val2) {
                    throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_WEIGHT))
                        ->compile('Transaction is overweight')->setData(compact('table'));
                }
            }
        }
        
        $commitResult = $this->doCommitTables($postProcessing, $clear);
        if (!isset($commitResult['fatal error'])) {
            Transaction::where(['uid' => $this->transactionData['UID']])->update(['status' => Transaction::STATUS_DONE]);
        }
        
        return $commitResult;
    }
    
    /**
     * Attempt to undo the transaction effects
     * 
     * @throws \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    public function undo()
    {
        if (self::isNotPrepared()) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_SEQUENCE))->compile('not prepared');
        }
        
        $this->tryReverse();
        app('session')->put(['transaction' => null]);
        Transaction::where(['uid' => $this->transactionData['UID']])->update(['status' => Transaction::STATUS_ABORT]);
    }
    
    /**
     * This method notifies Transactor about new supplied objects
     * 
     * @param bool $wasDirty
     * @param \Illuminate\Database\Eloquent\Model $model A Model instance
     * @throws \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    public static function receive(bool $wasDirty, $model)
    {
        $class = get_class($model);
        $writeTable = $model->getTable();
        $instance = self::getInstance();
        $table = $instance->getTableAllowedByClass($class);
        $keyName = $model->getKeyName();
        $instance->transactionData['received'][$table] += 1;
        
        if ($wasDirty) {
            $instance->transactionData['written'][$table] += 1;
            if ($writeTable == $instance->temporaryTables[$table]['insert']) {
                $instance->transactionData['new'][$table] += 1;
            }
        }
        
        if (!DTODefs::isUUIDable($table) && (new $class)->_incrementing) {
            $receivedAI = $instance->transactionData['oldAI'][$table] + $instance->transactionData['received'][$table];
            $limitAI = $instance->transactionData['newAI'][$table];
            if ($receivedAI > $limitAI) {
                throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_WEIGHT))
                    ->compile('Excessive objects supplied')->setData(compact('table', 'receivedAI', 'limitAI'));
            }
        }
        
        if ($keyName && isset($model->$keyName)) {
            LogReceivedRecord::create([
                'table' => $table,
                'key_name' => $keyName,
                'key' => $model->$keyName
            ]);
        }
        
        app('session')->put(['transaction' => $instance->transactionData]);
    }
    
    /**
     * This method gives a primary key for a new model
     * Do not increment the counter yet - what if 
     * 
     * @param \Illuminate\Database\Eloquent\Model|string $class A Model instance or a string representing its class
     */
    public static function nextId(string $class) {
        $instance = self::getInstance();
        $table = $instance->getTableAllowedByClass($class);
        
        if (DTODefs::isUUIDable($table)) {
            if (!isset($instance->uuidReserve[$table])) {
                if (!isset($instance->transactionData['new'][$table])) {
                    throw (new ApiTransactorException(ApiTransactorException::CODE_TABLE_NOT_INCLUDED))->compile($table);
                }

                $instance->uuidReserve[$table] = UUIDGeneration
                    ::where('table', $table)
                    ->orderBy('uuid', 'DESC')
                    ->skip($instance->transactionData['new'][$table])
                    ->take($instance->transactionData['reserved'][$table])
                    ->pluck('uuid')->toArray();
            }
            
            $uuid = array_shift($instance->uuidReserve[$table]);
            if (!isset($uuid)) {
                throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_WEIGHT))
                    ->compile('Excessive objects supplied')->setData(['table' => $table, 'uuids_num' => count($instance->uuidReserve[$table])]);
            }
            
            return $uuid;
        }
        
        return $instance->transactionData['oldAI'][$table] + $instance->transactionData['new'][$table];
    }
    
    /**
     * Prepare temporary tables
     * 
     * @param string $rootClass
     * @param array $tableCountData
     * @return null
     * @throws \App\Exceptions\ApiTransactorException
     */
    public static function prepare(string $rootClass, array $tableCountData)
    {
        if (self::isPrepared()) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_SEQUENCE))->compile('not closed');
        } else if (!self::isNotPrepared()) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_FAIL_NOT_UNDONE))->compile();
        }
        
        if (!is_a($test = new $rootClass, $rootClass) || !is_a($test, Model::class)) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_SERVER_FAULT))
                ->compile('Invalid argument: \$rootClass must be a valid class name of a class that extends Model')
                ->setData(compact('rootClass'));
        }
        
        $instance = self::getInstance();
        $instance->transactionData['UID'] = DB::selectOne('SELECT UUID_SHORT() as `uid`')->uid;
        
        if (!$instance->transactionData['UID']) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_UNABLE_TO_GENERATE_UID))->compile();
        }
        
        Transaction::create([
            'uid' => $instance->transactionData['UID'],
            'session_id' => $instance->sessionId
        ]);
        
        app('session')->put(['transaction' => $instance->transactionData]);
        
        $inspectionWorkLoad = ['auto-count' => &$tableCountData];
        $inspection = DTODefs::performInspection($rootClass, ['auto-count', 'enabled-write'], $inspectionWorkLoad);
        $instance->transactionData['dynUW'] = array_get($inspection, 'auto-count.dynUW', []);
        self::$allowedTransactionTables = &$inspection['enabled-write']['enabled'];
        self::$allowedTransactionTablesKeys = &$inspection['enabled-write']['key-name'];
        $sumCount = array_sum($tableCountData);
        
        LogReceivedRecord::establish(self::$allowedTransactionTablesKeys, $sumCount);
        self::$deletionEnabledTables = &$inspection['enabled-write']['deletable'];
        if (count(self::$deletionEnabledTables)) {
            DeletedTransactionRecord::establish(array_intersect_key(self::$allowedTransactionTablesKeys, self::$deletionEnabledTables), $sumCount);
        }
        
        if (count($instance->uuidables)) {
            UUIDGeneration::establish(array_keys(array_intersect_key(self::$allowedTransactionTablesKeys, $instance->uuidables)), $sumCount);
            
            foreach (array_intersect_key($tableCountData, self::$allowedTransactionTablesKeys, $instance->uuidables) as $table => $count) {
                if ($count > 0) {
                    DB::unprepared(UUIDGeneration::getGeneratorStatement($table, $count));
                }
            }
        }
        
        $instance->transactionData['rootClass'] = $rootClass;
        $instance->transactionData['isPrepared'] = 'abnormal';
        app('session')->put(['transaction' => $instance->transactionData]);
        
        foreach (self::$allowedTransactionTables as $modelClass => $table) {
            if ((new $modelClass)->_incrementing) {
                if (!isset($tableCountData[$table])) {
                    throw (new ApiTransactorException(ApiTransactorException::CODE_COUNT_MISSING))->compile($table);
                }

                $reserveCount = (int) $tableCountData[$table];

                if ($reserveCount == 0) {
                    continue;
                } else if ($reserveCount < 0) { // TODO: some maximum value to check?
                    throw (new ApiTransactorException(ApiTransactorException::CODE_COUNT_INVALID))->compile($table);
                }

                $instance->createTempTablesForModel($modelClass, $reserveCount);
            } else {
                $instance->createTempTablesForModel($modelClass);
            }
        }
        
        $instance->transactionData['isPrepared'] = true;
        app('session')->put(['transaction' => $instance->transactionData]);
    }
    
    /**
     * Clone the table structure into the new temporary tables for transaction,
     * reserve the specified amount of id's in the original table and
     * prepare the temporary tables for inserts to receive objects with these id's
     * 
     * @param string $modelClass A Model class
     * @param int|null $reserveCount
     * @return null
     * @throws \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    private function createTempTablesForModel($modelClass, int $reserveCount = null)
    {
        $table = (new $modelClass)->getTable();
        $isUuidable = DTODefs::isUUIDable($table);
        $this->showCreateParser->loadColumnData($table);
        $createStatementSplit = $this->showCreateParser->getCreateForTemp($table, $isUuidable);;
        $createTempTables = [];
        
        // first, create temporary tables
        foreach (['insert', 'update'] as $type) {
            $tempTable = $tempTables[$type] = $this->prefixTableForTransaction($table, $type);
            $createStatementSplit[0] = "CREATE TABLE `{$tempTable}` (";
            $createTempTables[] = implode("\n", $createStatementSplit);
            $this->temporaryTables[$table][$type] = $tempTable;
        }
        
        // bulk create of temporary tables
        if (!DB::unprepared(implode(";\n", $createTempTables))) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_CREATE_TABLE_FAIL))
                ->compile($type, $table);
        }
        
        // if necessary, preserve some A_I for possible new records
        if (isset($reserveCount)) {
            if ($isUuidable) {
                // just store the reserved amount
                $this->transactionData['reserved'][$table] = $reserveCount;
                $this->registerTableTransactionData($table);
            } else {
                // now, reserve amount of id's in original table and get its old AUTO_INCREMENT
                $oldAI = $this->changeAIOfRealTable($table, $reserveCount);

                // then, prerare temporary insert table for reserved id's
                if (!DB::statement("ALTER TABLE `{$tempTables['insert']}` AUTO_INCREMENT = {$oldAI}")) {
                    throw (new ApiTransactorException(ApiTransactorException::CODE_ALTER_TABLE_FAIL))
                        ->compile('change AUTO_INCREMENT', 'insert', $table);
                }

                // finally, remove the primary key auto_increment from update table
                if (!DB::statement("ALTER TABLE `{$tempTables['update']}` CHANGE `id` `id` INT(10) UNSIGNED NOT NULL")) {
                    throw (new ApiTransactorException(ApiTransactorException::CODE_ALTER_TABLE_FAIL))
                        ->compile('remove AUTO_INCREMENT', 'update', $table);
                }
            }
        } else {
            $this->registerTableTransactionData($table);
        }
    }
    
    /**
     * Reserve an amount of id's in the real table using change of AUTO_INCREMENT.
     * Return the old AUTO_INCREMENT
     * 
     * @param string $table
     * @param int $reserveCount
     * @return int
     * @throws \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    private function changeAIOfRealTable(string $table, int $reserveCount): int
    {        
        $status = DB::selectOne("SHOW TABLE STATUS WHERE `name` = '{$table}'");
        $oldAI = $status->Auto_increment;
        $newAI = $oldAI + $reserveCount;
        
        if (($success = is_int($oldAI) && DB::statement("ALTER TABLE `{$table}` AUTO_INCREMENT = {$newAI}"))) {
            // remember the state to be able to reverse if need
            $this->transactionData['oldAI'][$table] = $oldAI;
            $this->transactionData['newAI'][$table] = $newAI;
            $this->registerTableTransactionData($table);
        }
        
        if (!$success) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_ALTER_TABLE_FAIL))
                ->compile('change AUTO_INCREMENT', 'original', $table);
        }
        
        return $oldAI;
    }
    
    /**
     * Register basic Transaction data for the table
     * 
     * @param string $table
     * @return $this
     */
    private function registerTableTransactionData(string $table)
    {
        $this->transactionData['received'][$table] = 0;
        $this->transactionData['written'][$table] = 0;
        $this->transactionData['new'][$table] = 0;
        $this->transactionData['objects'] = [];
        app('session')->put(['transaction' => $this->transactionData]);
        
        return $this;
    }
    
    /**
     * Get tables from transactions of crashed or timed out sessions
     * 
     * @param \Illuminate\Database\Eloquent\Collection|array $clearableSessions
     * @return array
     */
    public static function tablesForGC($clearableSessions) {
        $database = env('DB_DATABASE');
        $tableNameField = "Tables_in_{$database}";

        $regex = self::transactionTablesRegex();
        $allTransactionTables = DB::select("SHOW TABLES WHERE `{$tableNameField}` REGEXP '{$regex}'");
        $outdatedTransactionTables = [];
        
        foreach ($allTransactionTables as $record) {
            $table = $record->$tableNameField;
            
            // a table name may have incomplete session id due to table name length limit
            $tableSession = ltrim(strrchr($table, '_'), '_');
            
            foreach ($clearableSessions as $uid) {
                if (strpos($uid, $tableSession) === 0) {
                    $outdatedTransactionTables[$uid][] = $table;
                    break;
                }
            }
        }
        
        foreach ($clearableSessions as $uid) {
            $supplementaries = [
                (new DeletedTransactionRecord([], $uid))->getTable(),
                (new UUIDGeneration([], $uid))->getTable(),
                (new LogReceivedRecord([], $uid))->getTable()
            ];
            
            foreach ($supplementaries as $table) {
                if (Schema::hasTable($table)) {
                    $outdatedTransactionTables[$uid][] = $table;
                }
            }
        }
        
        return $outdatedTransactionTables;
    }
    
    /**
     * Makes an attempt to reverse as possible of the failed transaction
     * If any of the original tables got new its AUTO_INCREMENT changed, it will not be reverted
     */
    private function tryReverse()
    {
        foreach ($this->temporaryTables as $table => $tempTable) {
            if (isset($this->transactionData['oldAI'][$table]) && isset($this->transactionData['newAI'][$table])) {
                try {
                    $status = DB::selectOne("SHOW TABLE STATUS WHERE `name` = '{$table}'");
                    
                    // Check the table have not got any new records and the AUTO_INCREMENT is same as we left it
                    if (isset($status->Auto_increment) && $status->Auto_increment == $this->transactionData['newAI'][$table]) {
                        // Revert it back to old one
                        DB::statement("ALTER TABLE `{$table}` AUTO_INCREMENT = {$this->transactionData['oldAI'][$table]}");
                    }
                } catch (\Throwable $ex) {}
            }
                
            try {
                DB::statement("DROP TABLE IF EXISTS `{$tempTable['insert']}`");
            } catch (\Throwable $ex) {}

            try {
                DB::statement("DROP TABLE IF EXISTS `{$tempTable['update']}`");
            } catch (\Throwable $ex) {}
        }
        
        UUIDGeneration::disband();
        LogReceivedRecord::disband();
        DeletedTransactionRecord::disband();
    }
    
    /**
     * Actual transaction finishing
     * 
     * @param \Closure|null $postProcessing
     * @param bool $clear
     * @return array empty if everything gone OK
     */
    private function doCommitTables(\Closure $postProcessing = null, bool $clear = true): array
    {
        $failedToCommit = [];
        $currentTable = null;
        $currentOp = null;
        $formerStatus = $this->transactionData;
        
        try {
            DB::transaction(function () use (&$currentTable, &$currentOp, $postProcessing) {
                $allowedTablesKeys = self::getAllowedTransactionTablesKeys();
                $allowedUuidables = array_intersect_key($allowedTablesKeys, $this->uuidables, $this->temporaryTables);
                $generationsTbl = (new UUIDGeneration)->getTable();
                $deletesTable = (new DeletedTransactionRecord)->getTable();
                $mentions = $this->showCreateParser->getUuidablesMentions($allowedTablesKeys);
                
                foreach ($allowedUuidables as $table => $keyName) {
                    $logsTable = (new LogReceivedRecord)->getTable();
                    DB::statement("SELECT * FROM `{$table}` WHERE `{$keyName}` > (SELECT MAX(`{$keyName}`) FROM `{$table}`) FOR UPDATE");
                    DB::statement("SET @ai = (SELECT `AUTO_INCREMENT` FROM `information_schema`.`tables` WHERE `table_name` = '{$table}' AND `table_schema` = DATABASE())");
                    DB::statement("UPDATE `{$generationsTbl}` SET `id` = (@ai:= @ai + 1) - 1 WHERE `table` = '{$table}'");
                }
                
                foreach ($mentions as $table => $keys) {
                    foreach (array_wrap($keys) as $key) {
                        foreach (['insert', 'update'] as $type) {
                            if (!isset($this->temporaryTables[$table][$type])) {
                                continue;
                            }

                            DB::statement($this->getUUIDUpdateStatement($this->temporaryTables[$table][$type], $generationsTbl, $key));
                        }
                    }
                }
                
                DB::statement($this->getUUIDUpdateStatement($deletesTable, $generationsTbl, 'key'));
                DB::statement($this->getUUIDUpdateStatement($logsTable, $generationsTbl, 'key'));
                
                if (count(self::$deletionEnabledTables)) {
                    $currentOp = 'groupsForDelete';
                    $groupsForDelete = DeletedTransactionRecord::groupBy('table')->pluck('key_name', 'table');
                    
                    foreach ($groupsForDelete as $table => $keyName) {
                        $currentOp = 'delete';
                        
                        DB::statement(
                            "DELETE `{$table}` FROM `{$table}`"
                            . " INNER JOIN `{$deletesTable}`"
                                . " ON `{$deletesTable}`.`table` = '{$table}'"
                                . " AND `{$deletesTable}`.`key` = `{$table}`.`{$keyName}`"
                        );
                    }
                }
                
                foreach ($this->temporaryTables as $table => $tempTable) {
                    $currentTable = $table;
                    foreach (['insert', 'update'] as $type) {
                        $currentOp = $type;
                        $this->doMoveTable($table, $tempTable[$type], $type);
                    }
                }
                
                self::stopWatchingTables();

                if (isset($postProcessing)) {
                    $currentOp = 'postProcess';
                    DB::statement("SET FOREIGN_KEY_CHECKS=1");
                    $postProcessing($this);
                    DB::statement("SET FOREIGN_KEY_CHECKS=0");
                }
            });
        } catch (\Throwable $exception) {
            $this->transactionData['isPrepared'] = 'abnormal';
            app('session')->put(['transaction' => $this->transactionData]);
            
            if ($currentOp === 'postProcess') {
                throw (new ApiTransactorException(ApiTransactorException::CODE_POST_PROCESS_FAILED, $exception))->compile();
            } else {
                throw (new ApiTransactorException(ApiTransactorException::CODE_FAIL_UNDONE, $exception))
                    ->compile($currentOp, $currentTable);
            }
        }
        
        if ($clear) {
            UUIDGeneration::disband();
            LogReceivedRecord::disband();
            DeletedTransactionRecord::disband();
            foreach ($this->temporaryTables as $table => $tempTable) {
                if (isset($this->transactionData['newAI'][$table])) {
                    try {
                        $statusInsert = DB::selectOne("SHOW TABLE STATUS WHERE `name` = '{$tempTable['insert']}'");
                        // Check if the insert table has less records than reserved amount (i.e. there was updates instead)
                        if (isset($statusInsert->Auto_increment) && $statusInsert->Auto_increment < $this->transactionData['newAI'][$table]) {
                            $status = DB::selectOne("SHOW TABLE STATUS WHERE `name` = '{$table}'");

                            // Check the table have not got any new records and the AUTO_INCREMENT is same as we left it
                            if (isset($status->Auto_increment) && $status->Auto_increment == $this->transactionData['newAI'][$table]) {
                                // Lets give back the difference
                                DB::statement("ALTER TABLE `{$table}` AUTO_INCREMENT = {$statusInsert->Auto_increment}");
                            }
                        }
                    } catch (\Throwable $exception) {
                        array_set($failedToCommit, "error.{$table}", [
                            'Cannot check and give back (if possible) AUTO_INCREMENT'
                        ]);

                        ExceptionDebug::makeDebug($failedToCommit["fatal error"], $exception, $table);
                    }
                }

                foreach (['insert', 'update'] as $type) {
                    DB::statement("DROP TABLE IF EXISTS `{$tempTable[$type]}`");
                }

                unset($this->transactionData['received'][$table]);
                unset($this->transactionData['written'][$table]);
                unset($this->transactionData['new'][$table]);
                unset($this->transactionData['oldAI'][$table]);
                unset($this->transactionData['newAI'][$table]);
                $this->transactionData['objects'] = [];
                
                app('session')->put(['transaction' => $this->transactionData]);
            }
        }

        try {
            $warning_messages = DB::select('SHOW WARNINGS');
            if (count($warning_messages)) {
                array_set($failedToCommit, "warning", $warning_messages);
            }
        } catch (\Throwable $exception) {
            array_set($failedToCommit, "warning", 'Failed to SHOW WARNINGS');
        }
        
        app('session')->put(['transaction' => null]);
        
        if (count($failedToCommit)) {
            $failedToCommit['status'] = $formerStatus;
        }

        return $failedToCommit;
    }
    
    /**
     * Update statement replacing UUIDs with auto-increments
     * 
     * @param string $tempTable
     * @param string $generationsTbl
     * @param string $key
     * @return string
     */
    private function getUUIDUpdateStatement(string $tempTable, string $generationsTbl, string $key)
    {
        return
            "UPDATE `{$tempTable}` as `temp_table`"
            . " INNER JOIN `{$generationsTbl}` as `gens_table` ON `gens_table`.`uuid` = `temp_table`.`{$key}`"
            . " SET `temp_table`.`{$key}` = `gens_table`.`id`;";
    }
    
    /**
     * Run inspection for allowed tables
     * 
     * @param array $workLoad
     * @param string $rootClass
     * @return $this
     */
    private function runInspections(array $workLoad = [], string $rootClass = null)
    {
        $this->inspectionResults = DTODefs::performInspection(
            $rootClass ?? $this->transactionData['rootClass'],
            ['preload', 'enabled-write', 'eagers'], $workLoad
        );
        
        self::$allowedTransactionTables = &$this->inspectionResults['enabled-write']['enabled'];
        self::$allowedTransactionTablesKeys = &$this->inspectionResults['enabled-write']['key-name'];
        self::$deletionEnabledTables = &$this->inspectionResults['enabled-write']['deletable'];
        
        return $this;
    }
    
    /**
     * REPLACE statement from transaction table
     * 
     * @param string $table
     * @param string $moveTable
     * @param srtring $type
     * @param int $shouldBeReplaced
     * @return null
     * @throw \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    private function doMoveTable($table, $moveTable, $type)
    {
        if ($type == 'insert') {
            $op = 'INSERT';
            $estimate = $this->transactionData['new'][$table];
        } else {
            $op = 'REPLACE';
            $estimate = 2 * ($this->transactionData['written'][$table] - $this->transactionData['new'][$table]);
        }
        
        // do not risk and waste time on nothing
        if ($estimate == 0) {
            return;
        }

        $affected = DB::affectingStatement(
            "{$op} INTO `{$table}` SELECT * FROM `{$moveTable}`"
        );

        if ($affected != $estimate) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_UNEXPECTED_CASUALTIES))
                ->compile($type, $affected, $op, $estimate);
        }
    }
    
    /**
     * Get regex to match any temporary table
     * 
     * @return string
     */
    private static function transactionTablesRegex(): string
    {
        $mergeTables = array_reduce(
            DTODefs::getGlobalRootDTOs(),
            function ($carry, $rootClass) {
                return $carry + self::getAllowedTransactionTables(false, $rootClass);
            }, []
        );
            
        $tablesOptions = implode('|', $mergeTables);
        $typesOptions = implode('|', ['insert', 'update']);
        $sessionIdChars = 'a-z0-9';
        return "({$tablesOptions})_({$typesOptions})_([{$sessionIdChars}]+)";
    }
    
    /**
     * Get table for a class
     * 
     * @param string $class
     * @return string
     * @throws \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    private function getTableAllowedByClass($class) : string
    {
        if (!self::isPrepared()) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_WRONG_SEQUENCE))->compile('not prepared');
        }
        
        $allowedTransactionModels = self::getAllowedTransactionTables();
        if (!isset($allowedTransactionModels[$class])) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_UNEXPECTED_CLASS))->compile($class);
        }
        
        return $allowedTransactionModels[$class];
    }
    
    /**
     * Prefixes table name for transaction table
     * 
     * @param string $tableName
     * @param string $type
     * @return string
     */
    private function prefixTableForTransaction(string $tableName, string $type): string
    {
        $ret = "{$tableName}_{$type}_{$this->transactionData['UID']}";
        
        // limit of table name length is 64
        if (strlen($ret) > 64) {
            return substr($ret, 0, 64);
        }
        
        return $ret;
    }
    
    /**
     * A singleton, no instantiation from outside
     */
    private function __construct()
    {
        $this->uuidables = DTODefs::getUUIDables();
        $this->sessionId = strtolower(\Illuminate\Support\Facades\Session::getId());
        $this->showCreateParser = ParseCreateTable::getInstance();
        $this->bufferedWriter = new BufferedWriter;
        $sessionData = app('session')->get('transaction');
        
        if ($sessionData && array_key_exists('UID', $sessionData)) {
            $transaction = Transaction::find($sessionData['UID']);
            
            if (!$sessionData['isPrepared'] || !$transaction || $transaction->status !== Transaction::STATUS_ACTIVE) {
                $transaction = null;
                $this->setDefaultTransactionData(true);
            } else {
                $this->transactionData = $sessionData;

                foreach (array_keys($sessionData['received']) as $tableName) {
                    $this->temporaryTables[$tableName] = [
                        'insert' => $this->prefixTableForTransaction($tableName, 'insert'),
                        'update' => $this->prefixTableForTransaction($tableName, 'update'),
                    ];
                }
                
                $transaction->touch();
            }
        } else {
            $this->setDefaultTransactionData(false);
        }
    }
    
    /**
     * Initial transaction data
     * 
     * @param bool $put
     * @return $this
     */
    private function setDefaultTransactionData(bool $put)
    {
        $this->transactionData = [
            'isPrepared' => 0,
            'received' => [],
            'written' => [],
            'new' => [],
            'oldAI' => [],
            'newAI' => [],
            'dynUW' => [],
            'objects' => [],
            'UID' => null
        ];
        
        if ($put) {
            app('session')->put(['transaction' => $this->transactionData]);
        }
        
        return $this;
    }
    
    /**
     * Get Transactor instance
     * 
     * @return \InSegment\ApiCore\Services\Transactor
     */
    private static function getInstance()
    {
        if (self::$instance == null) {
            self::$instance = new Transactor();
        }
        
        return self::$instance;
    }
    
}
